/*
 *  GTRunLoopDelegate.cpp
 *  All
 *
 *  Created by nhn on 11. 1. 1..
 *  Copyright 2011 nhn KEH. All rights reserved.
 *
 */
#include "StdAfx.h"
#include "GTEventDelegate_Android.h"
#include "GTLog.h"
#include "TypeDefinition.h"
#include <algorithm>
#include <errno.h>
#include "GTLog.h"
#include "BindMessageLooper.h"


static int g_timerHandle = 0;

#define INITLOCK(mutex, attr)	{ int _reti = pthread_mutex_init(&mutex, &attr);	ASSERT(_reti == 0); }
#define LOCK(mutex)				{ int _retl = pthread_mutex_lock(&mutex);			ASSERT(_retl == 0); }
#define TRYLOCK(mutex)			pthread_mutex_trylock(&mutex);
#define UNLOCK(mutex)			{ int _retu = pthread_mutex_unlock(&mutex);			ASSERT(_retu == 0); }
#define DESTROYLOCK(mutex)		{ int _retd = pthread_mutex_destroy(&mutex);		ASSERT(_retd == 0); }

//------------------------------------------------------
// Create Function

IGTEventDelegate *CreateEventDelegate(IGTEventDelegateEvent *pEvent)
{
	return new CGTEventDelegate(pEvent);
}
//------------------------------------------------------

static void OnLooperSourceEvent(int bindkey, int handle, int param1, int param2)
{
	//LOGE("FireEvent:%d,%d,%d,%d", bindkey,  handle,  param1,  param2);
	CGTEventDelegate *pEventDelegate = (CGTEventDelegate*)bindkey;
	if(pEventDelegate)
		pEventDelegate->_FireSourceEvent(handle, (void*)param1, (void*)param2);
}

static void OnLooperTimerEvent(int bindkey, int handle, int param1)
{
	//LOGE("FireTimerEvent:%d,%d,%d", bindkey,  handle,  param1);
	CGTEventDelegate *pEventDelegate = (CGTEventDelegate*)bindkey;
	if(pEventDelegate)
		pEventDelegate->_FireTimerEvent(handle, (void*)param1);
}

CGTEventDelegate::CGTEventDelegate(IGTEventDelegateEvent *pEvent) : 
	m_pDelegateImpl(NULL), 
	m_nRetainCount(0), 
	m_nSourceTargetCount(0), 
	m_nCheckReadyObserver(NOTHING), 
	m_nStatusMark(ACTIVE),
	m_bObserverFireDoing(false),
	m_pEvent(pEvent)
{
	pthread_mutexattr_t   mta;
	int rc = 0;
	rc = pthread_mutexattr_init(&mta);
	ASSERT(rc == 0);
	rc = pthread_mutexattr_settype(&mta, PTHREAD_MUTEX_RECURSIVE);
	ASSERT(rc == 0);

	INITLOCK(m_mutexSourceTarget, mta);
	INITLOCK(m_mutexObserverTemp, mta);
	INITLOCK(m_mutexObserver, mta);
	INITLOCK(m_mutexStatus, mta);
	
	// KEH_ANDROID_IMPL_PLZ

	_Retain();
	
	if(m_pEvent)
		m_pEvent->OnGTEventDelegateEvent_DidCreated(this);

	CBindMessageLooper::registerCallback(OnLooperSourceEvent, OnLooperTimerEvent);
}


CGTEventDelegate::~CGTEventDelegate(void)
{
	ASSERT(m_nStatusMark == DELETE);
	
	DLOG("[ENTER] CGTEventDelegate::~CGTEventDelegate(void) : %p", this);

	LOCK(m_mutexSourceTarget);
	
	listsourcetarget_t::iterator it = m_mapSourceTarget.begin();
	for(; it != m_mapSourceTarget.end(); it++)
	{
		listprivate_t *plist = it->second;
		ASSERT(plist);
		
		listprivate_t::iterator it2 = plist->begin();
		for(; it2 != plist->end(); it2++)
		{
			free(it2->second);
		}
		plist->clear();
		
		delete plist;
	}
	m_mapSourceTarget.clear();
	
	UNLOCK(m_mutexSourceTarget);

	DESTROYLOCK(m_mutexSourceTarget);
	DESTROYLOCK(m_mutexObserverTemp);
	DESTROYLOCK(m_mutexObserver);
	DESTROYLOCK(m_mutexStatus);
	
	DLOG("[LEAVE] CGTEventDelegate::~CGTEventDelegate(void) : %p", this);
}


// This Function is NOT THREAD SAFE!!!
void CGTEventDelegate::Delete(void)
{
	if(DELETE != m_nStatusMark)
	{
		DLOG(">> CGTEventDelegate::Delete %p, m_nStatusMark == false thread %p impl %p", this, pthread_self(), m_pDelegateImpl);

		m_nStatusMark = DELETE;

		_Release();

		return;
	}

	DLOG(">> CGTEventDelegate::Delete %p, m_nStatusMark == true %p", this, pthread_self());
}


void CGTEventDelegate::ReleaseAllSourcesAndTimers(void)
{
	BEGIN_BLOCK_THREAD_SAFETY;
	
	DLOG(">> CGTEventDelegate::ReleaseAllSourcesAndTimers %p thread %p impl %p", this, pthread_self(), m_pDelegateImpl);
	
	m_nStatusMark = RELEASE;
	
	END_BLOCK_THREAD_SAFETY;
	
	_Release(false);
}


void CGTEventDelegate::RegisterObserver(IGTRunLoopObserver *pObserver)
{
	BEGIN_BLOCK_THREAD_SAFETY;

	DLOG(">> [ENTER] CGTEventDelegate::RegisterObserver %p => %p cnt : %d %d\n",  this, pObserver, m_listObservers.size(), m_listObservers_Newbie.size());

	LOCK(m_mutexObserver);

	if(false == m_bObserverFireDoing)
	{		
		m_listObservers.remove(pObserver);
		m_listObservers.push_back(pObserver);
	}
	else
	{
		_PushToNewbieList(pObserver);
	}

	UNLOCK(m_mutexObserver);
	
	DLOG(">> [LEAVE] CGTEventDelegate::RegisterObserver %p => %p cnt : %d %d\n",  this, pObserver, m_listObservers.size(), m_listObservers_Newbie.size());
	
	END_BLOCK_THREAD_SAFETY;
}

void CGTEventDelegate::UnRegisterObserver(IGTRunLoopObserver *pObserver)
{
	BEGIN_BLOCK_THREAD_SAFETY;

	DLOG(">> [ENTER] CGTEventDelegate::UnRegisterObserver %p => %p cnt : %d %d\n",  this, pObserver, m_listObservers.size(), m_listObservers_Garbage.size());

	LOCK(m_mutexObserver);

	if(false == m_bObserverFireDoing)
	{
		m_listObservers.remove(pObserver);
	}
	else
	{
		_PushToGarbageList(pObserver);
	}

	UNLOCK(m_mutexObserver);
	
	DLOG(">> [LEAVE] CGTEventDelegate::UnRegisterObserver %p => %p cnt : %d %d\n",  this, pObserver, m_listObservers.size(), m_listObservers_Garbage.size());
	
	END_BLOCK_THREAD_SAFETY;
}


GT_EVENT_HANDLE CGTEventDelegate::SetTimer(GT_EVENT_HANDLE handle, int msec, GT_PARAM param)
{
	BEGIN_BLOCK_THREAD_SAFETY;

	if(handle == 0)
		handle = ::InterlockedIncrement(&g_timerHandle);

	settimerhandle_t::iterator it = m_setTimerHandle.find(handle);
	if(it == m_setTimerHandle.end())
	{
		_Retain();

		m_setTimerHandle.insert(handle);

		CBindMessageLooper::setTimer((int)this, handle, msec, (int)param);

		//GTLog::PrintCallStack(NULL, "", 0);
	}

	END_BLOCK_THREAD_SAFETY;
	

	return handle;
}


void CGTEventDelegate::KillTimer(GT_EVENT_HANDLE handle)
{
	BEGIN_BLOCK_THREAD_SAFETY;
		
	CBindMessageLooper::killTimer((int)this, handle);

	settimerhandle_t::iterator it = m_setTimerHandle.find(handle);
	if(it != m_setTimerHandle.end())
	{
		m_setTimerHandle.erase(it);
		_Release();
	}

	END_BLOCK_THREAD_SAFETY;
}


void CGTEventDelegate::RegisterSourceToCurrentRunLoop(int nPositionID, GT_EVENT_HANDLE handle)
{
	ASSERT(false);
}

void CGTEventDelegate::UnRegisterSourceToCurrentRunLoop(int nPositionID, GT_EVENT_HANDLE handle)
{
	ASSERT(false);
}


void CGTEventDelegate::PostSourceEventEx(int nPositionID, GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2)
{
	//LOGE("PostSourceEventEx: %d, %d, %d", handle, param1, param2);
	_Retain();
	CBindMessageLooper::postMessage((int)this, handle, (int)param1, (int)param2);
	//_FireSourceEvent(handle, param1, param2);
}


void CGTEventDelegate::PostSourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2)
{
	PostSourceEventEx(-1, handle, param1, param2);
}


void CGTEventDelegate::_PushToNewbieList(IGTRunLoopObserver *observer)
{
	LOCK(m_mutexObserverTemp);
	
	m_listObservers_Newbie.push_back(observer);
	
	if(m_listObservers_Newbie.size() > 0)
	{
		InterlockedExchange(&m_nCheckReadyObserver, EXIST);
	}
	
	UNLOCK(m_mutexObserverTemp);	
}


void CGTEventDelegate::_PushToGarbageList(IGTRunLoopObserver *observer)
{
	LOCK(m_mutexObserverTemp);
	
	m_listObservers_Garbage.push_back(observer);
	
	if(m_listObservers_Garbage.size() > 0)
	{
		InterlockedExchange(&m_nCheckReadyObserver, EXIST);
	}
	
	UNLOCK(m_mutexObserverTemp);
}


bool CGTEventDelegate::_FindGarbageList(IGTRunLoopObserver *observer)
{
	if(NOTHING == m_nCheckReadyObserver)
		return false;
	
	LOCK(m_mutexObserverTemp);

	listobserver_t::iterator it = std::find(m_listObservers_Garbage.begin(), m_listObservers_Garbage.end(), observer);
	
	UNLOCK(m_mutexObserverTemp);
	
	if(it != m_listObservers_Garbage.end())
		return true;
	
	return false;
}


void CGTEventDelegate::_ApplyObserverList()
{
	if(EXIST == m_nCheckReadyObserver)	// For Performance.. m_nCheckReadyObserver is interlocked operated..
	{
		DLOG(">> [ENTER] CGTEventDelegate::_Release() : m_nCheckReadyObserver == EXIST %p ==> %d %d",
				this, m_listObservers.size(), m_listObservers_Newbie.size(), m_listObservers_Garbage.size());
		
		LOCK(m_mutexObserver);
		LOCK(m_mutexObserverTemp);
		
		if(m_listObservers_Newbie.size() > 0)
		{
			listobserver_t::iterator it = m_listObservers_Newbie.begin();
			for (; it != m_listObservers_Newbie.end(); ++it)
			{
				m_listObservers.remove(*it);
				m_listObservers.push_back(*it);
			}
		}
		m_listObservers_Newbie.clear();
		
		
		if(m_listObservers_Garbage.size() > 0)
		{
			listobserver_t::iterator it = m_listObservers_Garbage.begin();
			for (; it != m_listObservers_Garbage.end(); ++it)
			{
				m_listObservers.remove(*it);
			}
		}	
		m_listObservers_Garbage.clear();
		
		UNLOCK(m_mutexObserverTemp);
		UNLOCK(m_mutexObserver);
		
		InterlockedExchange(&m_nCheckReadyObserver, NOTHING);
		
		DLOG(">> [LEAVE] CGTEventDelegate::_Release() : m_nCheckReadyObserver == EXIST %p ==> %d %d",
				this, m_listObservers.size(), m_listObservers_Newbie.size(), m_listObservers_Garbage.size());
	}
}


void CGTEventDelegate::_FireSourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2)
{	
	// Don't need to lock..
	// Already m_nRetainCount + 1 in onGTRunLoop_RegisterSource..
	// so you don't need to worry about that this class may be deleted while this function is processing.. 

	LOCK(m_mutexObserver);
	m_bObserverFireDoing = true;
	UNLOCK(m_mutexObserver);

	//LOGE("[ENTER] CGTEventDelegate::_FireSourceEvent : %p => handel = %d param1 = %d param2 = %d size = %d", this, handle, param1, param2, m_listObservers.size());
	
	listobserver_t::iterator it = m_listObservers.begin();
	for (; it != m_listObservers.end(); ++it)
	{
		if(_FindGarbageList(*it))
		{
			continue;
		}
		
		if((*it)->OnGTRunLoopObserver_SourceEvent(handle, param1, param2) == false)
			break;
	}
	//LOGE("[LEAVE] CGTEventDelegate::_FireSourceEvent : %p => handel = %d param1 = %d param2 = %d size = %d", this, handle, param1, param2, m_listObservers.size());
	
	// Don't need to lock by logic..
	m_bObserverFireDoing = false;

	_ApplyObserverList();
	
	_Release();
}


void CGTEventDelegate::_FireTimerEvent(GT_EVENT_HANDLE handle, GT_PARAM param)
{	
	// Don't need to lock..
	// Already m_nRetainCount + 1 in onGTRunLoop_RegisterTimer..
	// so you don't need to worry about that this class may be deleted while this function is processing.. 

	_Retain();
	
	LOCK(m_mutexObserver);
	m_bObserverFireDoing = true;
	UNLOCK(m_mutexObserver);
	
	//DLOG("[ENTER] CGTEventDelegate::_FireTimerEvent : %p => handel = %d param = %p size = %d", this, handle, param, m_listObservers.size());

	int i = 0;
	listobserver_t::iterator it = m_listObservers.begin();
	for (; it != m_listObservers.end(); ++it)
	{
		if(_FindGarbageList(*it))
		{
			continue;
		}
		
		if((*it)->OnGTRunLoopObserver_TimerEvent(handle, param) == false)
			break;	

		i++;
	}

	// Don't need to lock by logic..
	m_bObserverFireDoing = false;
	
	_ApplyObserverList();
	
	_Release();

	//DLOG("[LEAVE] CGTEventDelegate::_FireTimerEvent : %p => handel = %d param = %p size = %d", this, handle, param, m_listObservers.size());
}


